(***********************************************************************)
(*                                                                     *)
(*                               Grimage                               *)
(*                                                                     *)
(*                           Graphics image                            *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2014-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: colors.ml,v 1.9 2014/02/27 16:45:22 fclement Exp $ *)

(* Definitions and manipulations of colors. *)

type t = Graphics.color;;

type rgb_channel = int;;

type red = rgb_channel
and green = rgb_channel
and blue = rgb_channel
;;

type hsb_channel = int;;

type hue = hsb_channel
and saturation = hsb_channel
and brightness = hsb_channel
;;

(* Extra graphics functions primitives. *)
let nround x y = (2 * x + y) / (2 * y);;
let mxround x = nround (0xFF * x) 0xFF;;

(* color -> rgb. *)
let red_of_color c = (c lsr 16) land 0xFF;;
let green_of_color c = (c lsr 8) land 0xFF;;
let blue_of_color c = c land 0xFF;;

let r_of = red_of_color;;
let g_of = green_of_color;;
let b_of = blue_of_color;;

let rgb_of_color c = (r_of c, g_of c, b_of c);;

(* rgb -> color. *)
let color_of_rgb r g b = (r lsl 16) + (g lsl 8) + b;;
let rgb = color_of_rgb;;

(* rgb -> hsb. *)
let brightness_of_rgb r g b = max r (max g b);;

let get_min_and_dyn bri r g b =
  let min = min r (min g b) in
  min, bri - min
;;

let get_sat bri r g b =
  let _, d = get_min_and_dyn bri r g b in
  if bri = 0 then 0 else nround (0xFF * d) bri
;;

let saturation_of_rgb r g b =
  get_sat (brightness_of_rgb r g b) r g b
;;

let get_hue bri r g b =
  let min, d = get_min_and_dyn bri r g b in
  if bri = 0 || d = 0 then 0 else
  let r' = (bri - r) * 0xFF / d
  and g' = (bri - g) * 0xFF / d
  and b' = (bri - b) * 0xFF / d in
  let h =
    if bri = r then if min = g then 5 * 0xFF + b' else 0xFF - g' else
    if bri = g then if min = b then 0xFF + r' else 3 * 0xFF - b' else
    if min = r then 3 * 0xFF + g' else 5 * 0xFF - r' in
  h / 6
;;

let hue_of_rgb r g b =
 get_hue (brightness_of_rgb r g b) r g b
;;

let hsb_of_rgb r g b =
  let bri = brightness_of_rgb r g b in
  let sat = get_sat bri r g b in
  let hue = get_hue bri r g b in
  hue, sat, bri
;;

(* hsb -> rgb. *)
let rgb_of_hsb h s b =
  let h = h * 6 in
  let i = h / 0xFF * 0xFF in
  let f = h - i in
  let m = b * (0xFF - s) / 0xFF and n = b * (0xFF - s * f / 0xFF) / 0xFF
  and k = b * (0xFF - s * (0xFF - f) / 0xFF) / 0xFF in
  match i / 0xFF with
  | 0 | 6 -> mxround b, mxround k, mxround m
  | 1 -> mxround n, mxround b, mxround m
  | 2 -> mxround m, mxround b, mxround k
  | 3 -> mxround m, mxround n, mxround b
  | 4 -> mxround k, mxround m, mxround b
  | 5 -> mxround b, mxround m, mxround n
  | _ -> invalid_arg "rgb_of_hsb"
;;

(* hsb -> color. *)
let color_of_hsb h s b =
  let r, g, b = rgb_of_hsb h s b in
  color_of_rgb r g b
;;

let hsb = color_of_hsb;;

(* color -> hsb. *)
let hue_of_color c =
  let r, g, b = rgb_of_color c in
  hue_of_rgb r g b
;;

let saturation_of_color c =
  let r, g, b = rgb_of_color c in
  saturation_of_rgb r g b
;;

let brightness_of_color c =
  let r, g, b = rgb_of_color c in
  brightness_of_rgb r g b
;;

let hsb_of_color c =
  let r, g, b = rgb_of_color c in
  hsb_of_rgb r g b
;;

let hue_of = hue_of_color;;
let sat_of = saturation_of_color;;
let bri_of = brightness_of_color;;

